include "src/gb.inc"

; Global variables
SOAM equ $DE00
cur_keys equ $DF00
new_keys equ $DF01
nmis equ $DF02
oam_used = $DF03

player_xsub equ $DF10
player_x equ $DF11
player_dxsub equ $DF12
player_frame_sub equ $DF13
player_frame equ $DF14
player_facing equ $DF15


; Because the CPU runs without access to main RAM or ROM while OAM
; DMA is underway, an 11-byte subroutine needs to be kept in HRAM
; to keep the CPU busy until DMA is done.
run_dma equ $FFF4

; Instead of vectors, LR35902 has RSTs.
SECTION "rom", ROM0

  ; RST 00
  ret
  REPT 7
    nop
  ENDR
  ; RST 08
  ret
  REPT 7
    nop
  ENDR
  ; RST 10
  ret
  REPT 7
    nop
  ENDR
  ; RST 18
  ret
  REPT 7
    nop
  ENDR
  ; RST 20
  ret
  REPT 7
    nop
  ENDR
  ; RST 28
  ret
  REPT 7
    nop
  ENDR
  ; RST 30
  ret
  REPT 7
    nop
  ENDR
  ; RST 38
  ret
  REPT 7
    nop
  ENDR

; IRQs are RST 40-67, in order
; vblank, status (often hblank), timer, serial, keys

  jp draw
  REPT 5
    nop
  ENDR
  jp stat
  REPT 5
    nop
  ENDR
  jp timer
  REPT 5
    nop
  ENDR
  jp serial
  REPT 5
    nop
  ENDR
  jp joypad
  REPT 5
    nop
  ENDR

; Pad to the actual ROM header
  DS $100 - $68

  nop
  jp main

  ; magiccookiedata, used to ensure D7-D0 and A5-A0 work
  ; when deinterleaved to tiles, it spells "Nintendo"
  DB $CE,$ED,$66,$66,$CC,$0D,$00,$0B
  DB $03,$73,$00,$83,$00,$0C,$00,$0D
  DB $00,$08,$11,$1F,$88,$89,$00,$0E
  DB $DC,$CC,$6E,$E6,$DD,$DD,$D9,$99
  DB $BB,$BB,$67,$63,$6E,$0E,$EC,$CC
  DB $DD,$DC,$99,$9F,$BB,$B9,$33,$3E

  ; 11-character title
  DB "HELLO"
  DS 6

  ; Serial code (is ADME Doom or Animal Crossing?)
  DS 4

  ; Execution mode (bit 7 controls GBC switching)
  DB GBC_MONO

  ; $0144: Developer ID
  DB "OK"

  ; SGB mode
  DB SGB_NONE

  ; Mapper
  DB CART_ROM_ONLY

  ; $0148: ceil(log2(ROM size / 32768))
  DB $00
  
  ; RAM size: log2(RAM size / 512)/2
  DB $00

  ; Region code
  DB REGION_U

  ; Old hex developer ID: $33 means "see $0144"
  DB $33
  
  ; ROM version number
  DB $00

  ; Leave space for header checksum to be filled in by rgbfix
  DS 3

draw:
stat:
timer:
serial:
joypad:
    reti

main:
  di  ; Disable interrupts
  ld sp, RAM_END+1  ; Set up stack pointer (Empty descending)

  ; Reset the hardware
  call busy_wait_vblank  ; Wait for vblank
  xor a
  ld [LCDC], a  ; turn off rendering
  ld [SCY], a   ; clear scroll
  ld [SCX], a   ; clear scroll
  ld [NR52], a  ; disable audio

  ; Copy the sprite DMA delegate
  ld hl,run_dma_master
  ld de,run_dma
  ld bc,run_dma_end-run_dma_master
  call memcpy

  ; And this is where init.s would pass control to main.s

  call draw_bg
  call load_player_cels

  ; Clear OAM (is this necessary?)
  ld h, 0
  ld de, OAM_
  ld bc, 160
  call memset

  call busy_wait_vblank
  ; Set background palette 
  ld a,LCD_ON|BG_ON|OBJ_ON|BG_NT0|BG_CHR21
  ld [LCDC],a

.loop:
  call read_pad
  call move_player

  ; First make sure it doesn't crash
  xor a
  ld [oam_used],a
  call draw_player_sprite
  call lcd_clear_oam

  call wait_not_vblank
  call busy_wait_vblank
  call run_dma
  ld hl,nmis
  inc [hl]

  jr .loop

; ppuclear ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

lcd_off:
  ; Turning the screen off out of vertical blanking causes the
  ; display driver to get stuck on a scanline.
  ; (This stuck state is the same as the dark horizontal line when
  ; you turn off the Game Boy.)
  call busy_wait_vblank

  ; Use a RMW instruction to turn off only bit 7
  push hl
  ld hl, LCDC
  res 7, [hl]
  pop hl
  ret

lcd_on:
  ; If in vblank, turn on the display
  ld a, [LY]
  cp 144
  jr nc, .skip
    push hl
    ld hl, LCDC
    set 7, [hl]
    pop hl
.skip:
  ret

; Busy-wait for either forced blank or Y coordinate to be at least
; 145.  Use this before VRAM uploads or before changing LCDC bit 7.
busy_wait_vblank:
  ; If LCDC bit 7 already clear, we're already in forced blanking
  ld a,[LCDC]
  rlca
  ret nc

  ; Otherwise, wait for LCDC to become 145
.wait:
  ld a, [LY]
  cp 145
  jr c, .wait
  ret

; Busy-wait for either forced blank or Y coordinate to be at least
; 145.  Use this before VRAM uploads or before changing LCDC bit 7.
wait_not_vblank:
  ld a, [LY]
  cp 145
  jr nc, wait_not_vblank
  ret

memset:
  ld a, h
  ld [de],a
  inc de
  dec bc
  ld a, b
  or c
  jr nz, memset
  ret

memcpy:
  ld a, [hl+]
  ld [de],a
  inc de
  dec bc
  ld a, b
  or c
  jr nz, memcpy
  ret

; Background drawing ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

draw_bg:
  ; Fill CHR RAM
  ld hl,bggfx_chr
  ld de,CHRRAM2
  ld bc,16*96
  call memcpy
  
  ; Clear nametable
  ld h, 0
  ld de, NTRAM0
  ld bc, 1024
  call memset

  ; Draw top of floor
  ld h,$0B
  ld de,NTRAM0+32*16
  ld bc,20
  call memset

  ; Draw bottom of floor
  ld h,$01
  ld de,NTRAM0+32*17
  ld bc,20
  call memset
  
  ld hl,NTRAM0+32*12+1
  call bg_draw_one_block
  ld hl,NTRAM0+32*14+1
  call bg_draw_one_block
  ld hl,NTRAM0+32*12+17
  call bg_draw_one_block
  ld hl,NTRAM0+32*14+17
  call bg_draw_one_block

  ; Set palette
  ld a, %01101100
  ld [BGP],a
  ret  

bg_draw_one_block:
  ld a,$0C
  ld [hl+],a
  inc a
  ld [hl-],a
  inc a
  set 5,l
  ld [hl+],a
  inc a
  ld [hl-],a
  inc a
  ret

bggfx_chr: incbin "obj/gb/bggfx.chrgb"

; Controller reading ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

read_pad:
  ; Poll half the controller
  ld a,P1_BUTTONS
  call .onenibble
  ld b,a  ; B7-4 = 1; B3-0 = unpressed buttons

  ; Poll the other half
  ld a,P1_DPAD
  call .onenibble
  swap a   ; A3-0 = unpressed directions; A7-4 = 1
  xor b    ; A = pressed buttons + directions
  ld b,a   ; B = pressed buttons + directions

  ; And release the controller
  ld a,P1_NONE
  ld [P1],a

  ; Combine with previous cur_keys to make new_keys
  ld a,[cur_keys]
  xor b    ; A = keys that changed state
  and b    ; A = keys that changed to pressed
  ld [new_keys],a
  ld a,b
  ld [cur_keys],a
  ret

.onenibble
  ld [P1],a
  ld a,[P1]
  ld a,[P1]
  ld a,[P1]
  ld a,[P1]
  ld a,[P1]
  ld a,[P1]
  or $F0   ; A7-4 = 1; A3-0 = unpressed keys
  ret

; Player movement ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

WALK_SPD equ 105  ; speed limit in 1/256 px/frame
WALK_ACCEL equ 4  ; movement acceleration in 1/256 px/frame^2
WALK_BRAKE equ 8  ; stopping acceleration in 1/256 px/frame^2

LEFT_WALL equ 24
RIGHT_WALL equ 136

move_player:
  ld a,[cur_keys]
  ld e,a

  ; Accelerate to right only if the player is holding right
  ; on the Control Pad and has a nonnegative velocity.
  ld a,[player_dxsub]
  bit BKEY_RIGHT,e
  jr z,.notRight
  bit 7,a
  jr nz,.notRight
  
  ; Right is pressed. Add to velocity but don't allow it to exceed maximum.
  add WALK_ACCEL
  cp WALK_SPD
  jr c,.walk_speed_not_maxed
  ld a,WALK_SPD
.walk_speed_not_maxed:
  ld [player_dxsub],a
  ld hl,player_facing
  res BOAM_HFLIP,[hl]
  jr .doneRight
.notRight:

  ; Right is not pressed. Brake if headed right.  
  bit 7,a
  jr nz,.doneRight
  cp WALK_BRAKE
  jr nc,.notRightStop
  ld a,WALK_BRAKE
.notRightStop:
  sub WALK_BRAKE
  ld [player_dxsub],a
.doneRight:

  ; Accelerate to left only if the player is holding left
  ; on the Control Pad and has a nonpositive velocity.
  ld a,[player_dxsub]
  bit BKEY_LEFT,e
  jr z,.notLeft
  or a
  jr z,.isLeft
  bit 7,a
  jr z,.notLeft
.isLeft:
  ; Left is pressed.  Subtract from velocity.
  sub WALK_ACCEL
  cp 256-WALK_SPD
  jr nc,.walk_speed_not_neg_maxed
  ld a,256-WALK_SPD
.walk_speed_not_neg_maxed:
  ld [player_dxsub],a
  ld hl,player_facing
  set BOAM_HFLIP,[hl]
  jr .doneLeft

  ; Left is not pressed.  Brake if headed left.
.notLeft:
  bit 7,a
  jr z,.doneLeft
  cp 256-WALK_BRAKE
  jr c,.notLeftStop
  ld a,256-WALK_BRAKE
.notLeftStop:
  add WALK_BRAKE
  ld [player_dxsub],a

.doneLeft:

  ; Move the player by adding the velocity (in de) to the position
  ; (in hl)
  ld a,[player_dxsub]
  ld e,a
  ld d,0
  bit 7,a
  jr z,.player_dxlo_pos
  dec d
.player_dxlo_pos:
  ld a,[player_xsub]
  ld l,a
  ld a,[player_x]
  ld h,a
  add hl,de
  
  ; Test for collision with side walls
  ld a,h
  cp LEFT_WALL-4
  jr nc,.notHitLeft
  ld h,LEFT_WALL-4
  ld e,0
  jr .doneWallCollision
.notHitLeft:
  cp RIGHT_WALL-12
  jr c,.notHitRight
  ld h,RIGHT_WALL-13
  ld e,0
.notHitRight:
.doneWallCollision:

  ; Write back new velocity
  ld a,l
  ld [player_xsub],a
  ld a,h
  ld [player_x],a
  ld a,e
  ld [player_dxsub],a

  ; Animate the player
  ; If stopped, freeze the animation on frame 0
  or a
  jr nz, .notStop1
  ld a,$C0
  ld [player_frame_sub],a
  xor a
  jr .have_player_frame
.notStop1:

  ; Take absolute value of velocity (negate it if it's negative)
  bit 7,a
  jr z,.player_animate_noneg
  cpl
  inc a
.player_animate_noneg:

  ; Multiply abs(velocity) by 5/16
  sra a
  sra a
  ld b,a
  sra a
  sra a
  adc b
  ld b,a

  ; 16-bit add it to player_frame
  ld a,[player_frame_sub]
  add b
  ld [player_frame_sub],a
  ld a,[player_frame]
  adc 0

  ; Wrap from $800 (after last frame of walk cycle)
  ; to $100 (first frame of walk cycle)
  cp 8
  jr c,.have_player_frame
  ld a,1
.have_player_frame:

  ld [player_frame],a
  ret

; Sprite drawing ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

PLAYER_Y = 128

run_dma_master:
  ld a,SOAM >> 8
  ld [OAM_DMA],a
  ld a,40
.loop:
  dec a
  jr nz,.loop
  ret
run_dma_end:

lcd_clear_oam:
  ; Destination address in shadow OAM
  ld h,high(SOAM)
  ld a,[oam_used]
  and $FC
  ld l,a
  
  ; iteration count
  rrca
  rrca
  add 256 - 40
  ld c,a

  xor a
.rowloop:
  ld [hl+],a
  inc l
  inc l
  inc l
  inc c
  jr nz, .rowloop
  ret

load_player_cels:
  ld a, %00011110
  ld [OBP0],a
  ld a,40
  ld [player_x],a
  xor a
  ld [player_xsub],a
  ld [player_dxsub],a
  ld [player_facing],a

  ld hl,spritegfx_chr
  ld de,CHRRAM0+$100
  ld bc,16*48
  jp memcpy

draw_player_sprite:
Ldraw_y set $FF80
Lx_add set $FF82
Lrows_left set $FF84
Lrow_first_tile set $FF85
Ldraw_x_left set $FF87

  ld a,3
  ld [Lrows_left],a

  ; Platform games often use the bottom center
  ld a,PLAYER_Y
  sub 24
  add 16
  ld [Ldraw_y],a
  
  ; Set up increments based on flip value
  ld hl,player_facing
  ld c,[hl]  ; Bit 5 of player_facing is nz if facing left
  ld a,[player_x]
  ld b,8
  bit 5,c
  jr z,.not_flipped
  add 16-8
  ld b,256-8
.not_flipped:
  add a,8  ; Game Boy sprites start at X=8
  ld [Ldraw_x_left],a
  ld a,b
  ld [Lx_add],a

  ; The eight frames start at $10, $12, ..., $1E, where
  ; 0 is still and 1-7 are scooting.
  ld a,[player_frame]
  add a
  or $10
  ld [Lrow_first_tile],a

  ; Frame 7 is special: its hotspot is 1 unit forward.  Find which
  ; direction "forward" is.
  cp $1E
  jr c,.not_frame_7
  ld b,1
  bit 5,c
  jr z,.f7_not_flipped
  ld b,low(-1)
.f7_not_flipped:
  ld a,[Ldraw_x_left]
  add b
  ld [Ldraw_x_left],a
.not_frame_7:

  ld h,high(SOAM)
  ld a,[oam_used]
  ld l,a
.rowloop:
  ld b,2  ; B: remaining width on this row in 8px units
  ld a,[Lrow_first_tile]
  ld d,a  ; D: current tile
  ld a,[Ldraw_x_left]
  ld e,a  ; E: current X
.tileloop:
  ; Draw an 8x8 pixel chunk of the character using one 4-byte entry
  ; in the display list.
  ld a,[Ldraw_y]
  ld [hl+],a
  ld a,e  ; X position of sprite
  ld [hl+],a
  ld a,[Lx_add]
  add e
  ld e,a
  ld a,d  ; Tile number
  ld [hl+],a
  inc d
  ld a,c  ; Attribute
  ld [hl+],a
  dec b
  jr nz,.tileloop
  
  ; Move to the next row, which is 8 pixels down and on the next
  ; row of tiles in the pattern table.
  ld a,[Ldraw_y]
  add 8
  ld [Ldraw_y],a
  ld a,[Lrow_first_tile]
  add 16
  ld [Lrow_first_tile],a
  ld a,[Lrows_left]
  dec a
  ld [Lrows_left],a
  jr nz,.rowloop
  
  ld a,l
  ld [oam_used],a
  ret




spritegfx_chr: incbin "obj/gb/spritegfx.chrgb"
