include "src/gb.inc"
include "src/global.inc"

SECTION "ram_player", WRAM0
player_xsub: ds 1
player_x: ds 1
player_dxsub: ds 1
player_frame_sub: ds 1
player_frame: ds 1
player_facing: ds 1

SECTION "rom_player", ROM0

; Player movement ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

WALK_SPD equ 105  ; speed limit in 1/256 px/frame
WALK_ACCEL equ 4  ; movement acceleration in 1/256 px/frame^2
WALK_BRAKE equ 8  ; stopping acceleration in 1/256 px/frame^2

LEFT_WALL equ 24
RIGHT_WALL equ 136

move_player::
  ld a,[cur_keys]
  ld e,a

  ; Accelerate to right only if the player is holding right
  ; on the Control Pad and has a nonnegative velocity.
  ld a,[player_dxsub]
  bit BKEY_RIGHT,e
  jr z,.notRight
  bit 7,a
  jr nz,.notRight
  
  ; Right is pressed. Add to velocity but don't allow it to exceed maximum.
  add WALK_ACCEL
  cp WALK_SPD
  jr c,.walk_speed_not_maxed
  ld a,WALK_SPD
.walk_speed_not_maxed:
  ld [player_dxsub],a
  ld hl,player_facing
  res BOAM_HFLIP,[hl]
  jr .doneRight
.notRight:

  ; Right is not pressed. Brake if headed right.  
  bit 7,a
  jr nz,.doneRight
  cp WALK_BRAKE
  jr nc,.notRightStop
  ld a,WALK_BRAKE
.notRightStop:
  sub WALK_BRAKE
  ld [player_dxsub],a
.doneRight:

  ; Accelerate to left only if the player is holding left
  ; on the Control Pad and has a nonpositive velocity.
  ld a,[player_dxsub]
  bit BKEY_LEFT,e
  jr z,.notLeft
  or a
  jr z,.isLeft
  bit 7,a
  jr z,.notLeft
.isLeft:
  ; Left is pressed.  Subtract from velocity.
  sub WALK_ACCEL
  cp 256-WALK_SPD
  jr nc,.walk_speed_not_neg_maxed
  ld a,256-WALK_SPD
.walk_speed_not_neg_maxed:
  ld [player_dxsub],a
  ld hl,player_facing
  set BOAM_HFLIP,[hl]
  jr .doneLeft

  ; Left is not pressed.  Brake if headed left.
.notLeft:
  bit 7,a
  jr z,.doneLeft
  cp 256-WALK_BRAKE
  jr c,.notLeftStop
  ld a,256-WALK_BRAKE
.notLeftStop:
  add WALK_BRAKE
  ld [player_dxsub],a

.doneLeft:

  ; Move the player by adding the velocity (in de) to the position
  ; (in hl)
  ld a,[player_dxsub]
  ld e,a
  
  ; Idiom for sign extension, per ISSOtm in
  ; http://gbdev.gg8.se/forums/viewtopic.php?pid=3171#p3171
  ; which works because subtraction carry on 8080 is inverted
  ; compared to 6502/ARM (C true = -1)
  rla ; Bit 7 into carry
  sbc a, a ; $00 if no carry, $FF if carry
  ld d, a

  ld a,[player_xsub]
  ld l,a
  ld a,[player_x]
  ld h,a
  add hl,de
  
  ; Test for collision with side walls
  ld a,h
  cp LEFT_WALL-4
  jr nc,.notHitLeft
  ld h,LEFT_WALL-4
  ld e,0
  jr .doneWallCollision
.notHitLeft:
  cp RIGHT_WALL-12
  jr c,.notHitRight
  ld h,RIGHT_WALL-13
  ld e,0
.notHitRight:
.doneWallCollision:

  ; Write back new velocity
  ld a,l
  ld [player_xsub],a
  ld a,h
  ld [player_x],a
  ld a,e
  ld [player_dxsub],a

  ; Animate the player
  ; If stopped, freeze the animation on frame 0
  or a
  jr nz, .notStop1
  ld a,$C0
  ld [player_frame_sub],a
  xor a
  jr .have_player_frame
.notStop1:

  ; Take absolute value of velocity (negate it if it's negative)
  bit 7,a
  jr z,.player_animate_noneg
  cpl
  inc a
.player_animate_noneg:

  ; Multiply abs(velocity) by 5/16
  ; Consider alternative by ISSOtm in
  ; http://gbdev.gg8.se/forums/viewtopic.php?pid=3171#p3171
  sra a
  sra a
  ld b,a
  sra a
  sra a
  adc b
  ld b,a

  ; 16-bit add it to player_frame
  ld a,[player_frame_sub]
  add b
  ld [player_frame_sub],a
  ld a,[player_frame]
  adc 0

  ; Wrap from $800 (after last frame of walk cycle)
  ; to $100 (first frame of walk cycle)
  cp 8
  jr c,.have_player_frame
  ld a,1
.have_player_frame:

  ld [player_frame],a
  ret

; Sprite drawing ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

PLAYER_Y = 128
PLAYER_TILE_BASE = $50

;;
; Initializes the player character's position and copies its tiles
; into CHR RAM bank 0 starting at PLAYER_TILE_BASE.
load_player_cels::
  ld a, %00011110
  ldh [OBP0],a
  ld a,40
  ld [player_x],a
  xor a
  ld [player_xsub],a
  ld [player_dxsub],a
  ld [player_facing],a

  ld hl,spritegfx_chr
  ld de,CHRRAM0+PLAYER_TILE_BASE*16
  ld bc,16*48
  jp memcpy

;;
; Adds six sprites representing the player character
; to the display list.
draw_player_sprite::
  rsset hLocals
Ldraw_y         rb 1
Lx_add          rb 1
Lrows_left      rb 1
Lrow_first_tile rb 1
Ldraw_x_left    rb 1

  ld a,3
  ldh [Lrows_left],a

  ; Platform games often use the bottom center
  ld a,PLAYER_Y
  sub 24
  add 16
  ldh [Ldraw_y],a
  
  ; Set up increments based on flip value
  ld hl,player_facing
  ld c,[hl]  ; Bit 5 of player_facing is nz if facing left
  ld a,[player_x]
  ld b,8
  bit 5,c
  jr z,.not_flipped
  add 16-8
  ld b,256-8
.not_flipped:
  add a,8  ; Game Boy sprites start at X=8
  ldh [Ldraw_x_left],a
  ld a,b
  ldh [Lx_add],a

  ; The eight frames start at $10, $12, ..., $1E, where
  ; 0 is still and 1-7 are scooting.
  ld a,[player_frame]
  add a
  add PLAYER_TILE_BASE
  ldh [Lrow_first_tile],a

  ; Frame 7 is special: its hotspot is 1 unit forward.  Find which
  ; direction "forward" is.
  cp PLAYER_TILE_BASE + 7 * 2
  jr c,.not_frame_7
  ld b,1
  bit 5,c
  jr z,.f7_not_flipped
  ld b,low(-1)
.f7_not_flipped:
  ldh a,[Ldraw_x_left]
  add b
  ldh [Ldraw_x_left],a
.not_frame_7:

  ld h,high(SOAM)
  ld a,[oam_used]
  ld l,a
.rowloop:
  ld b,2  ; B: remaining width on this row in 8px units
  ldh a,[Lrow_first_tile]
  ld d,a  ; D: current tile
  ldh a,[Ldraw_x_left]
  ld e,a  ; E: current X
.tileloop:
  ; Draw an 8x8 pixel chunk of the character using one 4-byte entry
  ; in the display list.
  ldh a,[Ldraw_y]
  ld [hl+],a
  ld a,e  ; X position of sprite
  ld [hl+],a
  ldh a,[Lx_add]
  add e
  ld e,a
  ld a,d  ; Tile number
  ld [hl+],a
  inc d
  ld a,c  ; Attribute
  ld [hl+],a
  dec b
  jr nz,.tileloop
  
  ; Move to the next row, which is 8 pixels down and on the next
  ; row of tiles in the pattern table.
  ldh a,[Ldraw_y]
  add 8
  ldh [Ldraw_y],a
  ld a,[Lrow_first_tile]
  add 16
  ldh [Lrow_first_tile],a
  ldh a,[Lrows_left]
  dec a
  ldh [Lrows_left],a
  jr nz,.rowloop
  
  ld a,l
  ld [oam_used],a
  ret

spritegfx_chr: incbin "obj/gb/spritegfx.chrgb"
