include "src/gb.inc"
include "src/global.inc"

; The display list at $DE00-$DE9F gets DMA'd to OAM after every frame
; in which sprites moved.  Also called "shadow OAM".
section "ram_ppuclear",WRAM0[$DE00]
SOAM:: ds 160
oam_used:: ds 1  ; How much of the display list is used

SECTION "rom_ppuclear", ROM0

;;
; Waits for blanking and turns off rendering.
;
; Unlike NES and Super NES, which continuously generate a video
; signal to keep the TV's hsync and vsync circuits occupied, the
; Game Boy LCD uses freesync.  Thus the LCD driver halts entirely
; when rendering is off, not increasing LY.  Stopping the video
; signal outside vblank confuses the circuitry in the LCD panel,
; causing it to get stuck on a scanline.  This stuck state is the
; same as the dark horizontal line when you turn off the Game Boy.
;
; Turning rendering on, by contrast, can be done at any time and
; is done by writing the nametable base addresses and sprite size
; to LCDC with bit 7 set to true.
lcd_off::
  call busy_wait_vblank

  ; Use a RMW instruction to turn off only bit 7
  ld hl, LCDC
  res 7, [hl]
  ret

;;
; Waits for the vblank ISR to increment the count of vertical blanks.
; Will lock up if DI, vblank IRQ off, or LCD off.
; Clobbers A, HL
wait_vblank_irq::
  ld hl,nmis
  ld a,[hl]
.loop:
  halt
  cp [hl]
  jr z,.loop
  ret

;;
; Waits for forced blank (LCDC bit 7 clear) or vertical blank
; (LY >= 144).  Use before VRAM upload or before clearing LCDC bit 7.
busy_wait_vblank::
  ; If LCDC bit 7 already clear, we're already in forced blanking
  ldh a,[LCDC]
  rlca
  ret nc

  ; Otherwise, wait for LCDC to become 144 (not 145) because LY=0
  ; represents both the prerender line and the first visible line.
  ; This differs from Super NES, where the first visible line is 1,
  ; and NES, which has a post-render line numbered 240 between
  ; picture and vblank.
.wait:
  ldh a, [LY]
  cp 144
  jr c, .wait
  ret

  IF 0
;;
; Busy-wait for being out of vblank.  Use this for game loop timing
; if interrupts aren't in use yet.
wait_not_vblank::
  ldh a, [LY]
  cp 144
  jr nc, wait_not_vblank
  ret
  ENDC

;;
; Moves sprites in the display list from SOAM+[oam_used] through
; SOAM+$9C offscreen by setting their Y coordinate to 0, which is
; completely above the screen top (16).
lcd_clear_oam::
  ; Destination address in shadow OAM
  ld h,high(SOAM)
  ld a,[oam_used]
  and $FC
  ld l,a

  ; iteration count
  rrca
  rrca
  add 256 - 40
  ld c,a

  xor a
.rowloop:
  ld [hl+],a
  inc l
  inc l
  inc l
  inc c
  jr nz, .rowloop
  ret
